<?php

namespace FiloBlu\Esb\Model;

use DateTime;
use FiloBlu\Esb\Api\Data\QueueItemInterface;
use FiloBlu\Esb\Api\Data\QueueItemInterfaceFactory;
use FiloBlu\Esb\Api\QueueItemRepositoryInterface;
use Magento\Framework\Api\Filter;
use Magento\Framework\Api\SearchCriteriaInterface;
use Magento\Framework\Api\SearchResultsInterface;
use Magento\Framework\Api\SearchResultsInterfaceFactory;
use Magento\Framework\App\ResourceConnection;
use Magento\Framework\DB\Select;
use Magento\Framework\Exception\NoSuchEntityException;
use Zend_Db;
use const DATE_ATOM;

/**
 * Class QueueItemRespository
 * @package FiloBlu\Esb\Model
 */
class QueueItemRepository implements QueueItemRepositoryInterface
{
    /**
     * @var ResourceConnection
     */
    protected $resourceConnection;
    /**
     * @var QueueItemInterfaceFactory
     */
    private $queueItemFactory;
    /**
     * @var SearchResultsInterfaceFactory
     */
    private $searchResultsFactory;

    /**
     * QueueItemRespository constructor.
     * @param ResourceConnection $resourceConnection
     * @param QueueItemInterfaceFactory $queueItemFactory
     * @param SearchResultsInterfaceFactory $searchResultsFactory
     */
    public function __construct(ResourceConnection $resourceConnection,
                                QueueItemInterfaceFactory $queueItemFactory,
                                SearchResultsInterfaceFactory $searchResultsFactory)
    {
        $this->resourceConnection = $resourceConnection;
        $this->queueItemFactory = $queueItemFactory;
        $this->searchResultsFactory = $searchResultsFactory;
    }

    /**
     * @param $id
     * @return QueueItemInterface
     * @throws NoSuchEntityException
     */
    public function get($id)
    {
        $connection = $this->resourceConnection->getConnection();
        $table = $connection->getTableName('esb_schedule');

        $select = $connection->select()->from($table)->where('schedule_id = :id');

        $item = $connection->fetchRow($select, ['id' => $id], Zend_Db::FETCH_ASSOC);

        if (($item === null) || empty($item)) {
            throw new NoSuchEntityException(__("Could not find queue item with id = {$id}"));
        }

        return $this->fromData($item);
    }

    /**
     * @param array $item
     * @return QueueItemInterface
     */
    public function fromData(array $item)
    {
        $executableAfter = $item['executable_after'];

        $queueItem = $this->queueItemFactory->create();
        /** @var QueueItemInterface $queueItem */

        $queueItem->setId($item['schedule_id'])
            ->setEvent($item['event'])
            ->setPriority($item['priority'])
            ->setExecutableAfter(DateTime::createFromFormat(DATE_ATOM, $executableAfter));

        return $queueItem;

    }

    /**
     * @param QueueItemInterface $item
     * @return mixed
     */
    public function save(QueueItemInterface $item)
    {
        $connection = $this->resourceConnection->getConnection();
        $table = $connection->getTableName('esb_schedule');

        $bindings = [
            'event' => $item->getEvent(),
            'priority' => $item->getPriority(),
            'executable_after' => $item->getExecutableAfter() ? $item->getExecutableAfter()->format(DATE_ATOM) : null,
            'finished_at' => $item->getFinishedAt() ? $item->getFinishedAt()->format(DATE_ATOM) : null,
            'executed_at' => $item->getExecutedAt() ? $item->getExecutedAt()->format(DATE_ATOM) : null,
            'published_at' => $item->getPublishedAt() ? $item->getPublishedAt()->format(DATE_ATOM) : null,
        ];

        if ($item->getId()) {
            $connection->update($table, $bindings, $connection->quoteInto('schedule_id = ?', $item->getId()));

        } else {
            $connection->insert($table, $bindings);
            $item->setId($connection->lastInsertId($table));
        }
    }

    /**
     * @param QueueItemInterface $item
     * @return mixed
     */
    public function delete(QueueItemInterface $item)
    {
        if (!($id = $item->getId())) {
            return;
        }

        $connection = $this->resourceConnection->getConnection();
        $table = $connection->getTableName('esb_schedule');

        $connection->delete($table, ['schedule_id = ?' => $id]);
    }

    /**
     * @param SearchCriteriaInterface $searchCriteria
     * @return SearchResultsInterface
     */
    public function getList(SearchCriteriaInterface $searchCriteria)
    {
        $result = $this->searchResultsFactory->create();
        $result->setSearchCriteria($searchCriteria);

        $orders = $searchCriteria->getSortOrders();

        $connection = $this->resourceConnection->getConnection();
        $table = $connection->getTableName('esb_schedule');

        $select = $connection->select()->from($table, '*', 'main_table');

        foreach ($searchCriteria->getFilterGroups() as $filterGroup) {
            foreach ($filterGroup->getFilters() as $filter) {
                $this->getSqlCondition($select, $filter);
            }
        }

        foreach ($orders as $order) {
            $select->order("{$order->getField()} {$order->getDirection()}");
        }

        if ($searchCriteria->getPageSize() !== null) {

            if ($searchCriteria->getCurrentPage() === null) {
                $searchCriteria->setCurrentPage(1);
            }

            $select->limitPage(
                $searchCriteria->getCurrentPage(),
                $searchCriteria->getPageSize()
            );
        }
        $rows = $connection->fetchAll($select, [], Zend_Db::FETCH_ASSOC);

        $items = [];
        foreach ($rows as $row) {
            $items[] = $this->fromData($row);
        }

        $result->setItems($items);
        $result->setTotalCount(count($items));

        return $result;
    }

    /**
     * TODO: move in dedicated class
     * @see https://devdocs.magento.com/guides/v2.3/rest/performing-searches.html
     * @param Select $select
     * @param Filter $filter
     * @return string|void
     */
    protected function getSqlCondition(Select $select, Filter $filter)
    {
        switch ($filter->getConditionType()) {

            case 'from':
                return;
            case 'null':
                return $select->where("`{$filter->getField()}` IS NULL");
            case 'notnull':
                return $select->where("`{$filter->getField()}` IS NOT NULL");
            case 'gt':
                return $select->where("`{$filter->getField()}` > ?", $filter->getValue());
            case 'gteq':
                return $select->where("`{$filter->getField()}` >= ?", $filter->getValue());
            case 'lt':
                return $select->where("`{$filter->getField()}` < ?", $filter->getValue());
            case 'lteq':
                return $select->where("`{$filter->getField()}` <= ?", $filter->getValue());
            case 'like':
                return $select->where("`{$filter->getField()}` LIKE ?", $filter->getValue());
            case 'in':
                return $select->where("`{$filter->getField()}` IN (?)", $filter->getValue());
            case 'finset':
                return;
            case 'neq':
                return $select->where("`{$filter->getField()}` != ?", $filter->getValue());
            case 'eq':
            default:
                return $select->where("`{$filter->getField()}` = ?", $filter->getValue());
        }
    }
}